{-# LANGUAGE CPP, ForeignFunctionInterface, FlexibleContexts #-}
module BlueZSocket
    where

import Foreign
import Foreign.C.String
import Foreign.C.Types
import System.IO

import Data.Word
import BlueZDefs
import System.Posix.IO
import System.Posix.Types

import qualified System.Posix.Internals as Posix

#include "util.h"
#include <sys/socket.h>

rfSocket :: IO (BlueSock RFComm)
rfSocket = do
  fd <- {#call socket as _socket#} (fromIntegral afBluetooth) sockStream rfcommProto
  Posix.setNonBlockingFD fd
  return $ BlueSock fd

l2Socket :: IO (BlueSock L2Cap)
l2Socket = do
  fd <- {#call socket as _socket#} (fromIntegral afBluetooth) sockSeqpacket l2capProto
  Posix.setNonBlockingFD fd
  return $ BlueSock fd

bind :: Storable (BlueAddr a) => BlueSock a -> BlueAddr a -> IO ()
bind (BlueSock fd) addr = do 
  p <- fmap castPtr $ new addr 
  {#call bind as _bind#} fd p (fromIntegral $ sizeOf addr)
  return ()

listen :: BlueSock a -> Int -> IO ()
listen (BlueSock fd) i = do
  {#call listen as _listen#} fd (fromIntegral i) 
  return ()

accept' :: Storable (BlueAddr a) => BlueSock a -> IO (BlueSock a, BlueAddr a)
accept' (BlueSock fd) = do
  p <- malloc
  a <- peek p
  pint <- fmap castPtr $ new (sizeOf p)
  fd' <- {#call accept as _accept#} fd (castPtr p) pint
  addr <- peek p
  return (BlueSock fd', addr)

connect' :: (Storable (BlueAddr a)) => BlueSock a -> BlueAddr a -> IO ()
connect' (BlueSock fd) addr = do
  p <- fmap castPtr $ new addr
  {#call connect as _connect #} fd p (fromIntegral $ sizeOf addr)
  return ()

connect :: (Storable (BlueAddr a)) => BlueSock a -> BlueAddr a -> IO Handle
connect b@(BlueSock fd) addr = do
  connect' b addr
  fdToHandle (Fd fd)

{- very high level functions go here -}

connectToRFComm :: Address -> Int -> IO Handle
connectToRFComm addr chan = do
  chan' <- fmap fromIntegral $ {#call f_htobs#} (fromIntegral chan)
  sock <- rfSocket
  connect sock (BlueAddr (BluePort chan') addr)

connectToL2Cap :: Address -> Int -> IO Handle
connectToL2Cap addr psm = do
  psm' <- fmap fromIntegral $ {#call f_htobs#} (fromIntegral psm)
  sock <- l2Socket
  connect sock (BlueAddr (BluePort psm') addr)

listenOnRFComm :: Int -> IO (BlueSock RFComm)
listenOnRFComm chan = do
  sock <- rfSocket
  chan' <- fmap fromIntegral $ {#call f_htobs#} (fromIntegral chan)
  bind sock (BlueAddr (BluePort chan') bdAddrAny)
  listen sock 5
  return sock

listenOnL2Cap :: Int -> IO (BlueSock L2Cap)
listenOnL2Cap psm = do
  sock <- l2Socket
  psm' <- fmap fromIntegral $ {#call f_htobs#} (fromIntegral psm)
  bind sock (BlueAddr (BluePort psm') bdAddrAny)
  listen sock 5
  return sock

accept :: (Storable (BlueAddr a)) => BlueSock a ->  IO Handle
accept b = do
  (BlueSock fd,_) <- accept' b
  fdToHandle (Fd fd)

